@@ -9,10 +9,6 @@
 static const char hex_chars_lc[] = "0123456789abcdef";
 static const char hex_chars_uc[] = "0123456789ABCDEF";
 
-/**
- * init the buffer
- *
- */
 
 __attribute_noinline__
 buffer* buffer_init(void) {
@@ -56,7 +52,7 @@ void buffer_move(buffer * restrict b, buffer * restrict src) {
 /* make sure buffer is at least "size" big + 1 for '\0'. keep old data */
 __attribute_cold__
 __attribute_noinline__
-__attribute_nonnull__
+__attribute_nonnull__()
 __attribute_returns_nonnull__
 static char* buffer_realloc(buffer * const restrict b, const size_t len) {
     #define BUFFER_PIECE_SIZE 64uL  /*(must be power-of-2)*/
@@ -78,7 +74,7 @@ static char* buffer_realloc(buffer * const restrict b, const size_t len) {
 
 __attribute_cold__
 __attribute_noinline__
-__attribute_nonnull__
+__attribute_nonnull__()
 __attribute_returns_nonnull__
 static char* buffer_alloc_replace(buffer * const restrict b, const size_t size) {
     /*(discard old data so realloc() does not copy)*/
@@ -103,7 +99,7 @@ char* buffer_string_prepare_copy(buffer * const b, const size_t size) {
 
 __attribute_cold__
 __attribute_noinline__
-__attribute_nonnull__
+__attribute_nonnull__()
 __attribute_returns_nonnull__
 static char* buffer_string_prepare_append_resize(buffer * const restrict b, const size_t size) {
     if (b->used < 2) {  /* buffer_is_blank(b) */
@@ -300,20 +296,18 @@ void buffer_append_uint_hex_lc(buffer *b, uintmax_t value) {
 	}
 }
 
-__attribute_nonnull__
+__attribute_nonnull__()
 __attribute_returns_nonnull__
 static char* utostr(char buf[LI_ITOSTRING_LENGTH], uintmax_t val) {
 	char *cur = buf+LI_ITOSTRING_LENGTH;
+	uintmax_t x;
 	do {
-		int mod = val % 10;
-		val /= 10;
-		/* prepend digit mod */
-		*(--cur) = (char) ('0' + mod);
-	} while (0 != val);
+		*(--cur) = (char) ('0' + (int)(val - (x = val/10) * 10));
+	} while (0 != (val = x));           /* val % 10 */
 	return cur;
 }
 
-__attribute_nonnull__
+__attribute_nonnull__()
 __attribute_returns_nonnull__
 static char* itostr(char buf[LI_ITOSTRING_LENGTH], intmax_t val) {
 	/* absolute value not defined for INTMAX_MIN, but can take absolute
@@ -334,7 +328,6 @@ void buffer_append_int(buffer *b, intmax_t val) {
 }
 
 void buffer_append_strftime(buffer * const restrict b, const char * const restrict format, const struct tm * const restrict tm) {
-    force_assert(NULL != format);
     /*(localtime_r() or gmtime_r() producing tm should not have failed)*/
     if (__builtin_expect( (NULL == tm), 0)) return;
 
@@ -609,10 +602,6 @@ void buffer_append_string_encoded(buffer * const restrict b, const char * const
 	size_t d_len, ndx;
 	const char *map = NULL;
 
-	if (0 == s_len) return;
-
-	force_assert(NULL != s);
-
 	switch(encoding) {
 	case ENCODING_REL_URI:
 		map = encoded_chars_rel_uri;
@@ -628,8 +617,6 @@ void buffer_append_string_encoded(buffer * const restrict b, const char * const
 		break;
 	}
 
-	force_assert(NULL != map);
-
 	/* count to-be-encoded-characters */
 	for (ds = (unsigned char *)s, d_len = 0, ndx = 0; ndx < s_len; ds++, ndx++) {
 		if (map[*ds & 0xFF]) {
@@ -680,12 +667,89 @@ void buffer_append_string_encoded(buffer * const restrict b, const char * const
 	}
 }
 
+void buffer_append_string_encoded_json(buffer * const restrict b, const char * const restrict s, const size_t len) {
+	const unsigned char * const restrict ds = (unsigned char *)s;
+	size_t dlen = 0;
+
+	/* calculate space needed for string including encodings */
+	for (size_t i = 0; i < len; ++i) {
+		int c = ds[i];
+		if (c == '"' || c == '\\' || c < 0x20 || c == 0x7f) {
+			switch (c) {
+			  case '\b':
+			  case '\t':
+			  case '\n':
+			  case '\f':
+			  case '\r':
+			  case '"':
+			  case '\\':
+				dlen += 2;
+				break;
+			  default:
+				dlen += 6; /* \uCCCC */
+				break;
+			}
+		}
+		else {
+			++dlen;
+		}
+	}
+
+	unsigned char * const d = (unsigned char *)buffer_extend(b, dlen);
+
+	if (__builtin_expect( (dlen == len), 1)) {/*(short-circuit; nothing to encode)*/
+		memcpy(d, ds, len);
+		return;
+	}
+
+	dlen = 0;
+	for (size_t i = 0; i < len; ++i) {
+		int c = ds[i];
+		if (c == '"' || c == '\\' || c < 0x20 || c == 0x7f) {
+			d[dlen++] = '\\';
+			switch (c) {
+			  case '\b':
+				d[dlen++] = 'b';
+				break;
+			  case '\t':
+				d[dlen++] = 't';
+				break;
+			  case '\n':
+				d[dlen++] = 'n';
+				break;
+			  case '\f':
+				d[dlen++] = 'f';
+				break;
+			  case '\r':
+				d[dlen++] = 'r';
+				break;
+			  case '"':
+				d[dlen++] = '"';
+				break;
+			  case '\\':
+				d[dlen++] = '\\';
+				break;
+			  default:
+				d[dlen  ] = 'u';
+				d[dlen+1] = '0';
+				d[dlen+2] = '0';
+				d[dlen+3] = hex_chars_lc[(c >> 4) & 0x0F];
+				d[dlen+4] = hex_chars_lc[c & 0x0F];
+				dlen += 5;
+				break;
+			}
+		}
+		else {
+			d[dlen++] = c;
+		}
+	}
+}
+
+
 void buffer_append_string_c_escaped(buffer * const restrict b, const char * const restrict s, size_t s_len) {
 	unsigned char *ds, *d;
 	size_t d_len, ndx;
 
-	if (0 == s_len) return;
-
 	/* count to-be-encoded-characters */
 	for (ds = (unsigned char *)s, d_len = 0, ndx = 0; ndx < s_len; ds++, ndx++) {
 		if ((*ds < 0x20) /* control character */
@@ -753,9 +817,9 @@ void buffer_urldecode_path(buffer * const b) {
     char *dst = src;
     do {
         /* *src == '%' */
-        unsigned char high = hex2int(*(src + 1));
-        unsigned char low = hex2int(*(src + 2));
-        if (0xFF != high && 0xFF != low) {
+        unsigned char high = ((unsigned char *)src)[1];
+        unsigned char low = high ? hex2int(((unsigned char *)src)[2]) : 0xFF;
+        if (0xFF != (high = hex2int(high)) && 0xFF != low) {
             high = (high << 4) | low;   /* map ctrls to '_' */
             *dst = (high >= 32 && high != 127) ? high : '_';
             src += 2;
@@ -768,6 +832,7 @@ void buffer_urldecode_path(buffer * const b) {
 
 int buffer_is_valid_UTF8(const buffer *b) {
     /* https://www.w3.org/International/questions/qa-forms-utf-8 */
+    /*assert(b->used);*//*(b->ptr must exist and be '\0'-terminated)*/
     const unsigned char *c = (unsigned char *)b->ptr;
     while (*c) {
 
@@ -910,7 +975,8 @@ void buffer_path_simplify(buffer *b)
 
 void buffer_to_lower(buffer * const b) {
     unsigned char * const restrict s = (unsigned char *)b->ptr;
-    for (uint32_t i = 0; i < b->used; ++i) {
+    const uint_fast32_t used = b->used;
+    for (uint_fast32_t i = 0; i < used; ++i) {
         if (light_isupper(s[i])) s[i] |= 0x20;
     }
 }
@@ -918,7 +984,8 @@ void buffer_to_lower(buffer * const b) {
 
 void buffer_to_upper(buffer * const b) {
     unsigned char * const restrict s = (unsigned char *)b->ptr;
-    for (uint32_t i = 0; i < b->used; ++i) {
+    const uint_fast32_t used = b->used;
+    for (uint_fast32_t i = 0; i < used; ++i) {
         if (light_islower(s[i])) s[i] &= 0xdf;
     }
 }
