@@ -371,7 +371,7 @@ static chunk * chunkqueue_append_mem_chunk(chunkqueue *cq, size_t sz) {
     return c;
 }
 
-__attribute_nonnull__
+__attribute_nonnull__()
 __attribute_returns_nonnull__
 static chunk * chunkqueue_append_file_chunk(chunkqueue * const restrict cq, const buffer * const restrict fn, off_t offset, off_t len) {
     chunk * const c = chunk_acquire_filechunk();
@@ -506,9 +506,6 @@ void chunkqueue_append_buffer_commit(chunkqueue *cq) {
 }
 
 
-static void chunkqueue_remove_empty_chunks(chunkqueue *cq);
-
-
 char * chunkqueue_get_memory(chunkqueue * const restrict cq, size_t * const restrict len) {
 	size_t sz = *len ? *len : (chunk_buf_sz >> 1);
 	buffer *b;
@@ -573,6 +570,7 @@ void chunkqueue_set_tempdirs(chunkqueue * const restrict cq, const array * const
     cq->tempdir_idx = 0;
 }
 
+__attribute_noinline__
 static void chunkqueue_dup_file_chunk_fd (chunk * const restrict d, const chunk * const restrict c) {
     /*assert(d != c);*/
     /*assert(d->type == FILE_CHUNK);*/
@@ -589,55 +587,48 @@ static void chunkqueue_dup_file_chunk_fd (chunk * const restrict d, const chunk
     }
 }
 
+__attribute_noinline__
 static void chunkqueue_steal_partial_file_chunk(chunkqueue * const restrict dest, const chunk * const restrict c, const off_t len) {
     chunkqueue_append_file(dest, c->mem, c->offset, len);
     chunkqueue_dup_file_chunk_fd(dest->last, c);
 }
 
 void chunkqueue_steal(chunkqueue * const restrict dest, chunkqueue * const restrict src, off_t len) {
-	while (len > 0) {
-		chunk *c = src->first;
-		off_t clen = 0, use;
-
-		if (NULL == c) break;
+	for (off_t clen; len > 0; len -= clen) {
+		chunk * const c = src->first;
+		if (__builtin_expect( (NULL == c), 0)) break;
 
 		clen = chunk_remaining_length(c);
-		if (0 == clen) {
-			/* drop empty chunk */
-			src->first = c->next;
-			if (c == src->last) src->last = NULL;
-			chunk_release(c);
-			continue;
-		}
 
-		use = len >= clen ? clen : len;
-		len -= use;
-
-		if (use == clen) {
+		if (len >= clen) {
 			/* move complete chunk */
 			src->first = c->next;
 			if (c == src->last) src->last = NULL;
 
-			chunkqueue_append_chunk(dest, c);
-			dest->bytes_in += use;
+			if (__builtin_expect( (0 != clen), 1)) {
+				chunkqueue_append_chunk(dest, c);
+				dest->bytes_in += clen;
+			}
+			else /* drop empty chunk */
+				chunk_release(c);
 		} else {
-			/* partial chunk with length "use" */
+			/* copy partial chunk */
 
 			switch (c->type) {
 			case MEM_CHUNK:
-				chunkqueue_append_mem(dest, c->mem->ptr + c->offset, use);
+				chunkqueue_append_mem(dest, c->mem->ptr + c->offset, len);
 				break;
 			case FILE_CHUNK:
 				/* tempfile flag is in "last" chunk after the split */
-				chunkqueue_steal_partial_file_chunk(dest, c, use);
+				chunkqueue_steal_partial_file_chunk(dest, c, len);
 				break;
 			}
 
-			c->offset += use;
-			force_assert(0 == len);
+			c->offset += len;
+			clen = len;
 		}
 
-		src->bytes_out += use;
+		src->bytes_out += clen;
 	}
 }
 
@@ -814,6 +805,106 @@ int chunkqueue_append_mem_to_tempfile(chunkqueue * const restrict dest, const ch
 	return -1;
 }
 
+#ifdef HAVE_PWRITEV
+
+#ifdef HAVE_SYS_UIO_H
+#include <sys/uio.h>
+#endif
+
+__attribute_cold__
+__attribute_noinline__
+static ssize_t chunkqueue_append_cqmem_to_tempfile_partial(chunkqueue * const dest, chunk * const c, ssize_t wr, log_error_st * const restrict errh) {
+    /* recover from partial write of existing dest MEM_CHUNK to tempfile */
+    chunk *ckpt = dest->first;
+    while (ckpt->next != c) ckpt = ckpt->next;
+    ckpt->next = NULL;
+    dest->last = ckpt;
+    dest->bytes_in  -= wr; /*(avoid double count in dest cq)*/
+    dest->bytes_out -= wr;
+    chunkqueue_mark_written(dest, wr);/*(remove MEM_CHUNK written to tempfile)*/
+
+    c->next = dest->first; /*(place tempfile at beginning of dest cq)*/
+    dest->first = c;
+    return (0 == chunkqueue_to_tempfiles(dest, errh)) ? 0 : -1;
+}
+
+static ssize_t chunkqueue_append_cqmem_to_tempfile(chunkqueue * const restrict dest, chunkqueue * const restrict src, off_t len, log_error_st * const restrict errh) {
+    /* write multiple MEM_CHUNKs to tempfile in single pwritev() syscall */
+    /*(could lseek() and writev() if pwritev() is not available,
+     * but if writev() is available, pwritev() is likely available,
+     * e.g. any modern Linux or *BSD, and possibly anything not Windows)*/
+    unsigned int iovcnt = 0;
+    struct iovec iov[16];
+
+    off_t dlen = 0;
+    chunk *c;
+    for (c = dest->first; c && c->type == MEM_CHUNK; c = c->next) {
+        const off_t clen = chunk_remaining_length(c);
+        iov[iovcnt].iov_base = c->mem->ptr + c->offset;
+        iov[iovcnt].iov_len  = (size_t)clen;
+        dlen += clen;
+        ++iovcnt;
+        if (__builtin_expect( (iovcnt == sizeof(iov)/sizeof(*iov)), 0))
+            break; /*(not expecting large number of MEM_CHUNK)*/
+    }
+    if (__builtin_expect( (c != NULL), 0) && dest->first->type == MEM_CHUNK) {
+        /*(expecting only MEM_CHUNK if dest cq starts w/ MEM_CHUNK)*/
+        /*(use less efficient fallback if that assumption does not hold true)*/
+        if (0 != chunkqueue_to_tempfiles(dest, errh))
+            return -1;
+        dlen = 0;
+        iovcnt = 0;
+    }
+
+    if (__builtin_expect( (iovcnt < sizeof(iov)/sizeof(*iov)), 1)) {
+        for (c = src->first; c && c->type == MEM_CHUNK; c = c->next) {
+            off_t clen = chunk_remaining_length(c);
+            if (clen > len) clen = len;
+            iov[iovcnt].iov_base = c->mem->ptr + c->offset;
+            iov[iovcnt].iov_len  = (size_t)clen;
+            len -= clen;
+            ++iovcnt;
+            if (0 == len) break;
+            if (__builtin_expect( (iovcnt == sizeof(iov)/sizeof(*iov)), 0))
+                break; /*(not expecting large number of MEM_CHUNK)*/
+        }
+    }
+
+    if (__builtin_expect( (0 == iovcnt), 0)) return 0; /*(should not happen)*/
+
+    c = chunkqueue_get_append_tempfile(dest, errh);
+    if (NULL == c)
+        return -1;
+  #ifdef __COVERITY__
+    if (c->file.fd < 0) return -1;
+  #endif
+    /* coverity[negative_returns : FALSE] */
+    ssize_t wr = pwritev(c->file.fd, iov, (int)iovcnt, c->file.length);
+
+    /*(memory use in chunkqueues is expected to be limited before spilling
+     * to tempfiles, so common case will write entire iovec to tempfile,
+     * and we return amount written *from src cq*, even if partial write;
+     * (not looping here to retry writing more, but caller might loop))*/
+
+    if (wr >= 0) {
+        c->file.length += wr;
+        dest->bytes_in += wr;
+        if (dlen) {
+            if (__builtin_expect( (wr < dlen), 0))
+                return
+                  chunkqueue_append_cqmem_to_tempfile_partial(dest,c,wr,errh);
+            wr -= (ssize_t)dlen;
+            dest->bytes_in  -= dlen; /*(avoid double count in dest cq)*/
+            dest->bytes_out -= dlen;
+            chunkqueue_mark_written(dest, dlen);
+        }
+    }
+
+    return wr;
+}
+
+#endif /* HAVE_PWRITEV */
+
 #ifdef HAVE_SPLICE
 
 __attribute_cold__
@@ -963,14 +1054,27 @@ ssize_t chunkqueue_append_splice_sock_tempfile(chunkqueue * const restrict cq, c
 #endif /* HAVE_SPLICE */
 
 int chunkqueue_steal_with_tempfiles(chunkqueue * const restrict dest, chunkqueue * const restrict src, off_t len, log_error_st * const restrict errh) {
-	while (len > 0) {
-		chunk *c = src->first;
-		off_t clen = 0, use;
+	for (off_t clen; len > 0; len -= clen) {
+		chunk * const c = src->first;
+		if (__builtin_expect( (NULL == c), 0)) break;
 
-		if (NULL == c) break;
+	  #ifdef HAVE_PWRITEV
+
+		if (c->type == MEM_CHUNK) {
+			clen = chunkqueue_append_cqmem_to_tempfile(dest, src, len, errh);
+			if (__builtin_expect( (clen < 0), 0)) return -1;
+			chunkqueue_mark_written(src, clen); /*(updates src->bytes_out)*/
+		}
+		else { /* (c->type == FILE_CHUNK) */
+			clen = chunk_remaining_length(c);
+			if (len < clen) clen = len;
+			chunkqueue_steal(dest, src, clen);/*(same as below for FILE_CHUNK)*/
+		}
+
+	  #else
 
 		clen = chunk_remaining_length(c);
-		if (0 == clen) {
+		if (__builtin_expect( (0 == clen), 0)) {
 			/* drop empty chunk */
 			src->first = c->next;
 			if (c == src->last) src->last = NULL;
@@ -978,46 +1082,46 @@ int chunkqueue_steal_with_tempfiles(chunkqueue * const restrict dest, chunkqueue
 			continue;
 		}
 
-		use = (len >= clen) ? clen : len;
-		len -= use;
-
 		switch (c->type) {
 		case FILE_CHUNK:
-			if (use == clen) {
+			if (len >= clen) {
 				/* move complete chunk */
 				src->first = c->next;
 				if (c == src->last) src->last = NULL;
 				chunkqueue_append_chunk(dest, c);
-				dest->bytes_in += use;
+				dest->bytes_in += clen;
 			} else {
-				/* partial chunk with length "use" */
+				/* copy partial chunk */
 				/* tempfile flag is in "last" chunk after the split */
-				chunkqueue_steal_partial_file_chunk(dest, c, use);
-				c->offset += use;
-				force_assert(0 == len);
+				chunkqueue_steal_partial_file_chunk(dest, c, len);
+				c->offset += len;
+				clen = len;
 			}
 			break;
 
 		case MEM_CHUNK:
-			/* store "use" bytes from memory chunk in tempfile */
-			if (0 != chunkqueue_append_mem_to_tempfile(dest, c->mem->ptr + c->offset, use, errh)) {
+			/* store bytes from memory chunk in tempfile */
+			if (0 != chunkqueue_append_mem_to_tempfile(dest, c->mem->ptr + c->offset,
+			                                           (len >= clen) ? clen : len, errh)) {
 				return -1;
 			}
 
-			if (use == clen) {
+			if (len >= clen) {
 				/* finished chunk */
 				src->first = c->next;
 				if (c == src->last) src->last = NULL;
 				chunk_release(c);
 			} else {
 				/* partial chunk */
-				c->offset += use;
-				force_assert(0 == len);
+				c->offset += len;
+				clen = len;
 			}
 			break;
 		}
 
-		src->bytes_out += use;
+		src->bytes_out += clen;
+
+	  #endif
 	}
 
 	return 0;
@@ -1055,24 +1159,21 @@ void chunkqueue_append_cq_range (chunkqueue * const dst, const chunkqueue * cons
 void chunkqueue_mark_written(chunkqueue *cq, off_t len) {
     cq->bytes_out += len;
 
-    for (chunk *c; (c = cq->first); ) {
+    for (chunk *c = cq->first; c; ) {
         off_t c_len = chunk_remaining_length(c);
         if (len >= c_len) { /* chunk got finished */
+            chunk * const x = c;
+            c = c->next;
             len -= c_len;
-            cq->first = c->next;
-            chunk_release(c);
-            if (0 == len) break;
+            chunk_release(x);
         }
         else { /* partial chunk */
             c->offset += len;
+            cq->first = c;
             return; /* chunk not finished */
         }
     }
-
-    if (NULL == cq->first)
-        cq->last = NULL;
-    else
-        chunkqueue_remove_finished_chunks(cq);
+    cq->first = cq->last = NULL;
 }
 
 void chunkqueue_remove_finished_chunks(chunkqueue *cq) {
@@ -1082,7 +1183,7 @@ void chunkqueue_remove_finished_chunks(chunkqueue *cq) {
     }
 }
 
-static void chunkqueue_remove_empty_chunks(chunkqueue *cq) {
+void chunkqueue_remove_empty_chunks(chunkqueue *cq) {
 	chunk *c;
 	chunkqueue_remove_finished_chunks(cq);
 
@@ -1156,9 +1257,6 @@ static int chunk_open_file_chunk(chunk * const restrict c, log_error_st * const
 	/*(skip file size checks if file is temp file created by lighttpd)*/
 	if (c->file.is_temp) return 0;
 
-	force_assert(FILE_CHUNK == c->type);
-	force_assert(c->offset >= 0 && c->offset <= c->file.length);
-
 	struct stat st;
 	if (-1 == fstat(c->file.fd, &st)) {
 		log_perror(errh, __FILE__, __LINE__, "fstat failed");
@@ -1167,7 +1265,8 @@ static int chunk_open_file_chunk(chunk * const restrict c, log_error_st * const
 
 	const off_t offset = c->offset;
 	const off_t len = c->file.length - c->offset;
-	if (offset > st.st_size || len > st.st_size || offset > st.st_size - len) {
+	force_assert(offset >= 0 && len >= 0);
+	if (offset > st.st_size - len) {
 		log_error(errh, __FILE__, __LINE__, "file shrunk: %s", c->mem->ptr);
 		return -1;
 	}
@@ -1357,6 +1456,8 @@ void
 chunkqueue_small_resp_optim (chunkqueue * const restrict cq)
 {
     /*(caller must verify response is small (and non-empty) before calling)*/
+    /*(caller must verify first chunk is MEM_CHUNK, i.e. response headers)*/
+    /*(caller must verify response is non-zero length)*/
 
     /*(optimization to use fewer syscalls to send a small response by reading
      * small files into memory, thereby avoiding use of sendfile() and multiple
@@ -1364,49 +1465,53 @@ chunkqueue_small_resp_optim (chunkqueue * const restrict cq)
      *(If TLS, then will shortly need to be in memory for encryption anyway)*/
 
     /*assert(cq->first);*/
-    chunk *c = cq->first;
-    chunk * const filec = c->next;
-    if (c->type != MEM_CHUNK || filec != cq->last || filec->type != FILE_CHUNK)
+    /*assert(cq->first->type == MEM_CHUNK);*/
+    /*assert(cq->first->next);*/
+    chunk * restrict c = cq->first;
+    chunk * const restrict filec = c->next;  /*(require file already be open)*/
+    if (filec != cq->last || filec->type != FILE_CHUNK || filec->file.fd < 0)
         return;
 
-    const int fd = filec->file.fd;
-    if (fd < 0) return; /*(require that file already be open)*/
-    off_t offset = filec->offset;
   #ifndef HAVE_PREAD
-    if (-1 == lseek(fd, offset, SEEK_SET)) return;
+    if (-1 == lseek(filec->file.fd, filec->offset, SEEK_SET)) return;
   #endif
 
     /* Note: there should be no size change in chunkqueue,
      * so cq->bytes_in and cq->bytes_out should not be modified */
 
-    buffer *b = c->mem;
     off_t len = filec->file.length - filec->offset;
-    if ((size_t)len > buffer_string_space(b)) {
-        chunk * const nc = chunk_acquire((size_t)len+1);
-        c->next = nc;
-        nc->next = filec;
-        b = nc->mem;
+    if ((size_t)len > buffer_string_space(c->mem)) {
+        c->next = chunk_acquire((size_t)len+1);
+        c = c->next;
+        /*c->next = filec;*/
     }
+    /* detach filec from chunkqueue; file expected to be read fully */
+    c->next = NULL;
+    cq->last = c;
 
-    char * const ptr = b->ptr + buffer_clen(b);
     ssize_t rd;
-  #ifdef HAVE_PREAD
-    const off_t foff = offset;
-  #endif
-    offset = 0; /*(reuse offset var for offset into mem buffer)*/
+    off_t offset = 0;
+    char * const ptr = buffer_extend(c->mem, len);
     do {
       #ifdef HAVE_PREAD
-        rd =pread(fd, ptr+offset, (size_t)len, foff+offset);
+        rd =pread(filec->file.fd, ptr+offset, (size_t)len,filec->offset+offset);
       #else
-        rd = read(fd, ptr+offset, (size_t)len);
+        rd = read(filec->file.fd, ptr+offset, (size_t)len);
       #endif
     } while (rd > 0 ? (offset += rd, len -= rd) : errno == EINTR);
     /*(contents of chunkqueue kept valid even if error reading from file)*/
-    if (len)
-        cq->bytes_in -= len;
-    buffer_commit(b, offset);
-    filec->offset += offset;
-    chunkqueue_remove_empty_chunks(cq);
+    if (__builtin_expect( (0 == len), 1))
+        chunk_release(filec);
+    else { /*(unexpected; error recovery)*/
+        buffer_truncate(c->mem, (uint32_t)(ptr + offset - c->mem->ptr));
+        cq->last = c->next = filec;
+        if (offset)
+            filec->offset += offset;
+        else if (__builtin_expect( (cq->first != c), 0)) {
+            cq->first->next = filec;
+            chunk_release(c);
+        }
+    }
 }
 
 
