@@ -39,6 +39,13 @@ void config_get_config_cond_info(config_cond_info * const cfginfo, uint32_t idx)
     cfginfo->comp_key = dc->comp_key;
 }
 
+int config_capture(server *srv, int idx) {
+    data_config * const dc = (data_config *)config_reference.data[idx];
+    return (dc->capture_idx)
+      ? dc->capture_idx
+      : (dc->capture_idx = ++srv->config_captures);
+}
+
 int config_feature_bool (const server *srv, const char *feature, int default_value) {
     return srv->srvconf.feature_flags
       ? config_plugin_value_tobool(
@@ -386,7 +393,9 @@ static cond_result_t config_check_cond_cached(request_st * const r, const data_c
     return config_check_cond_nocache_calc(r, dc, debug_cond, cache);
 }
 
-static int data_config_pcre_exec(const data_config *dc, cond_cache_t *cache, const buffer *b, cond_match_t *cond_match);
+static int config_pcre_match(request_st *r, const data_config *dc, const buffer *b);
+
+static cond_result_t config_check_cond_nocache_eval(request_st * const r, const data_config * const dc, const int debug_cond, cond_cache_t * const cache);
 
 static cond_result_t config_check_cond_nocache(request_st * const r, const data_config * const dc, const int debug_cond, cond_cache_t * const cache) {
 	/* check parent first */
@@ -460,6 +469,10 @@ static cond_result_t config_check_cond_nocache(request_st * const r, const data_
 		return (cache->local_result = COND_RESULT_TRUE);
 		/* remember result of local condition for a partial reset */
 
+	return config_check_cond_nocache_eval(r, dc, debug_cond, cache);
+}
+
+static cond_result_t config_check_cond_nocache_eval(request_st * const r, const data_config * const dc, const int debug_cond, cond_cache_t * const cache) {
 	/* pass the rules */
 
 	static struct const_char_buffer {
@@ -542,14 +555,13 @@ static cond_result_t config_check_cond_nocache(request_st * const r, const data_
 		match ^= (buffer_is_equal(l, &dc->string));
 		break;
 	case CONFIG_COND_NOMATCH:
-	case CONFIG_COND_MATCH: {
-		cond_match_t *cond_match = r->cond_match + dc->context_ndx;
+	case CONFIG_COND_MATCH:
 		match = (dc->cond == CONFIG_COND_MATCH);
-		match ^= (data_config_pcre_exec(dc, cache, l, cond_match) > 0);
+		match ^= (config_pcre_match(r, dc, l) > 0);
 		break;
-	}
 	default:
 		match = 1; /* return (cache->local_result = COND_RESULT_FALSE); below */
+		break;
 	}
 	/* remember result of local condition for a partial reset */
 	cache->local_result = match ? COND_RESULT_FALSE : COND_RESULT_TRUE;
@@ -596,8 +608,6 @@ static void config_cond_clear_node(cond_cache_t * const cond_cache, const data_c
 
 /**
  * reset the config-cache for a named item
- *
- * if the item is COND_LAST_ELEMENT we reset all items
  */
 void config_cond_cache_reset_item(request_st * const r, comp_key_t item) {
 	cond_cache_t * const cond_cache = r->cond_cache;
@@ -626,26 +636,77 @@ void config_cond_cache_reset(request_st * const r) {
 		memset(r->cond_cache, 0, used*sizeof(cond_cache_t));
 }
 
-#ifdef HAVE_PCRE_H
+#ifdef HAVE_PCRE2_H
+#define PCRE2_CODE_UNIT_WIDTH 8
+#include <pcre2.h>
+#elif defined(HAVE_PCRE_H)
 #include <pcre.h>
 #endif
 
-static int data_config_pcre_exec(const data_config *dc, cond_cache_t *cache, const buffer *b, cond_match_t *cond_match) {
-#ifdef HAVE_PCRE_H
-    #ifndef elementsof
-    #define elementsof(x) (sizeof(x) / sizeof(x[0]))
-    #endif
-    cache->patterncount =
+static int config_pcre_match(request_st * const r, const data_config * const dc, const buffer * const b) {
+
+  #ifdef HAVE_PCRE2_H
+
+    if (__builtin_expect( (0 == dc->capture_idx), 1))
+        return pcre2_match(dc->code, (PCRE2_SPTR)BUF_PTR_LEN(b),
+                           0, 0, dc->match_data, NULL);
+
+    const int capture_offset = dc->capture_idx - 1;
+    cond_match_t * const cond_match =
+      r->cond_match[capture_offset] = r->cond_match_data + capture_offset;
+    pcre2_match_data *match_data = cond_match->match_data;
+    if (__builtin_expect( (NULL == match_data), 0)) {
+        /*(allocate on demand)*/
+      #if 0 /*(if we did not want to share dc->match_data across requests)*/
+        /* index 0 is reused for all matches for which captures not used by
+         * other directives within the condition, so allocate for up to 9
+         * captures, plus 1 for %0 for full match.  Number of captures is
+         * checked at startup to be <= 9 in data_config_pcre_compile()
+         * (future: could save a few bytes if max captures were calculated
+         *  at startup in config_finalize()) */
+        match_data = cond_match->match_data = (0 == dc->capture_idx)
+          ? pcre2_match_data_create(10, NULL)
+          : pcre2_match_data_create_from_pattern(dc->code, NULL);
+      #else
+        match_data = cond_match->match_data =
+          pcre2_match_data_create_from_pattern(dc->code, NULL);
+      #endif
+        force_assert(match_data);
+        cond_match->matches = pcre2_get_ovector_pointer(match_data);
+    }
+    cond_match->comp_value = b; /*holds pointer to b (!) for pattern subst*/
+    cond_match->captures =
+      pcre2_match(dc->code, (PCRE2_SPTR)BUF_PTR_LEN(b), 0, 0, match_data, NULL);
+    return cond_match->captures;
+
+  #elif defined(HAVE_PCRE_H)
+
+    if (__builtin_expect( (0 == dc->capture_idx), 1)) {
+        int matches[3 * 10];
+        return pcre_exec(dc->regex, dc->regex_study, BUF_PTR_LEN(b), 0, 0,
+                         matches, sizeof(matches)/sizeof(*matches));
+    }
+
+    const int capture_offset = dc->capture_idx - 1;
+    cond_match_t * const cond_match =
+      r->cond_match[capture_offset] = r->cond_match_data + capture_offset;
+    if (__builtin_expect( (NULL == cond_match->matches), 0)) {
+        /*(allocate on demand)*/
+        cond_match->matches = malloc(dc->ovec_nelts * sizeof(int));
+        force_assert(cond_match->matches);
+    }
+    cond_match->comp_value = b; /*holds pointer to b (!) for pattern subst*/
+    cond_match->captures =
       pcre_exec(dc->regex, dc->regex_study, BUF_PTR_LEN(b), 0, 0,
-                cond_match->matches, elementsof(cond_match->matches));
-    if (cache->patterncount > 0)
-        cond_match->comp_value = b; /*holds pointer to b (!) for pattern subst*/
-    return cache->patterncount;
-#else
+                cond_match->matches, dc->ovec_nelts);
+    return cond_match->captures;
+
+  #else
+
+    UNUSED(r);
     UNUSED(dc);
-    UNUSED(cache);
     UNUSED(b);
-    UNUSED(cond_match);
     return 0;
-#endif
+
+  #endif
 }
