@@ -61,7 +61,7 @@ static data_unset *configparser_get_variable(config_t *ctx, const buffer *key) {
     fprintf(stderr, "get var on block: %s\n", dc->key.ptr);
     array_print(dc->value, 0);
 #endif
-    if (NULL != (du = array_get_element_klen(dc->value, CONST_BUF_LEN(key)))) {
+    if (NULL != (du = array_get_element_klen(dc->value, BUF_PTR_LEN(key)))) {
       data_unset *du_copy = du->fn->copy(du);
       buffer_clear(&du_copy->key);
       return du_copy;
@@ -81,7 +81,7 @@ static data_unset *configparser_merge_data(data_unset *op1, const data_unset *op
       buffer_append_int(&ds->value, ((data_integer*)op2)->value);
       return op1;
     } else if (op1->type == TYPE_INTEGER && op2->type == TYPE_STRING) {
-      data_string *ds = data_string_init();
+      data_string *ds = array_data_string_init();
       buffer_append_int(&ds->value, ((data_integer*)op1)->value);
       buffer_append_string_buffer(&ds->value, &((data_string*)op2)->value);
       op1->fn->free(op1);
@@ -103,58 +103,41 @@ static data_unset *configparser_merge_data(data_unset *op1, const data_unset *op
     case TYPE_ARRAY: {
       array *dst = &((data_array *)op1)->value;
       array *src = &((data_array *)op2)->value;
-      data_unset *du;
+      const data_unset *du, *ddu;
       size_t i;
 
       for (i = 0; i < src->used; i ++) {
         du = (data_unset *)src->data[i];
         if (du) {
-          if (buffer_is_empty(&du->key) || !array_get_element_klen(dst, CONST_BUF_LEN(&du->key))) {
+          if (buffer_is_unset(&du->key)
+              || !(ddu = array_get_element_klen(dst, BUF_PTR_LEN(&du->key)))){
             array_insert_unique(dst, du->fn->copy(du));
           } else {
             fprintf(stderr, "Duplicate array-key '%s'\n", du->key.ptr);
+            if (ddu->type == du->type) {
+              /*(ignore if new key/value pair matches existing key/value)*/
+              if (du->type == TYPE_STRING
+                  && buffer_is_equal(&((data_string *)du)->value,
+                                     &((data_string *)ddu)->value))
+                  continue;
+              if (du->type == TYPE_INTEGER
+                  && ((data_integer*)du)->value == ((data_integer*)ddu)->value)
+                  continue;
+            }
             op1->fn->free(op1);
             return NULL;
           }
         }
       }
       break;
+    }
     default:
       force_assert(0);
       break;
-    }
   }
   return op1;
 }
 
-static int configparser_remoteip_normalize_compat(buffer *rvalue) {
-  /* $HTTP["remoteip"] IPv6 accepted with or without '[]' for config compat
-   * http_request_host_normalize() expects IPv6 with '[]',
-   * and config processing at runtime expects COMP_HTTP_REMOTE_IP
-   * compared without '[]', so strip '[]' after normalization */
-  buffer *b = buffer_init();
-  int rc;
-
-  if (rvalue->ptr[0] != '[') {
-      buffer_append_string_len(b, CONST_STR_LEN("["));
-      buffer_append_string_buffer(b, rvalue);
-      buffer_append_string_len(b, CONST_STR_LEN("]"));
-  } else {
-      buffer_append_string_buffer(b, rvalue);
-  }
-
-  rc = http_request_host_normalize(b, 0);
-
-  if (0 == rc) {
-    /* remove surrounding '[]' */
-    size_t blen = buffer_string_length(b);
-    if (blen > 1) buffer_copy_string_len(rvalue, b->ptr+1, blen-2);
-  }
-
-  buffer_free(b);
-  return rc;
-}
-
 __attribute_pure__
 static comp_key_t
 configparser_comp_key_id(const buffer * const obj_tag, const buffer * const comp_tag)
@@ -211,23 +194,27 @@ configparser_parse_condition(config_t * const ctx, const buffer * const obj_tag,
       return; /* unreachable */
     }
 
+    const uint32_t comp_offset = buffer_clen(&ctx->current->key)+3;
     buffer * const tb = ctx->srv->tmp_buf;
-    buffer_copy_buffer(tb, &ctx->current->key);
-    buffer_append_string_len(tb, CONST_STR_LEN(" / "));
-    const uint32_t comp_offset = buffer_string_length(tb);
-    buffer_append_string_len(tb, CONST_STR_LEN("$"));
-    buffer_append_string_buffer(tb, obj_tag); /*(HTTP, REQUEST_HEADER, SERVER)*/
-    buffer_append_string_len(tb, CONST_STR_LEN("[\""));
-    buffer_append_string_buffer(tb, comp_tag);
-    buffer_append_string_len(tb, CONST_STR_LEN("\"] "));
-    buffer_append_string_len(tb, op, 2);
-    buffer_append_string_len(tb, CONST_STR_LEN(" \""));
-    buffer_append_string_buffer(tb, rvalue);
-    buffer_append_string_len(tb, CONST_STR_LEN("\""));
+    buffer_clear(tb);
+    struct const_iovec iov[] = {
+      { BUF_PTR_LEN(&ctx->current->key) }
+     ,{ CONST_STR_LEN(" / ") }   /* comp_offset */
+     ,{ CONST_STR_LEN("$") }
+     ,{ BUF_PTR_LEN(obj_tag) } /*(HTTP, REQUEST_HEADER, SERVER)*/
+     ,{ CONST_STR_LEN("[\"") }
+     ,{ BUF_PTR_LEN(comp_tag) }
+     ,{ CONST_STR_LEN("\"] ") }
+     ,{ op, 2 }
+     ,{ CONST_STR_LEN(" \"") }
+     ,{ BUF_PTR_LEN(rvalue) }
+     ,{ CONST_STR_LEN("\"") }
+    };
+    buffer_append_iovec(tb, iov, sizeof(iov)/sizeof(*iov));
 
     data_config *dc;
     if (NULL != (dc = configparser_get_data_config(ctx->all_configs,
-                                                   CONST_BUF_LEN(tb)))) {
+                                                   BUF_PTR_LEN(tb)))) {
       configparser_push(ctx, dc, 0);
     }
     else {
@@ -253,40 +240,9 @@ configparser_parse_condition(config_t * const ctx, const buffer * const obj_tag,
       }
       else if (COMP_HTTP_REMOTE_IP == dc->comp
                && (dc->cond == CONFIG_COND_EQ || dc->cond == CONFIG_COND_NE)) {
-        char * const slash = strchr(rvalue->ptr, '/'); /* CIDR mask */
-        char * const colon = strchr(rvalue->ptr, ':'); /* IPv6 */
-        if (NULL != slash && slash == rvalue->ptr){/*(skip AF_UNIX /path/file)*/
-        }
-        else if (NULL != slash) {
-          char *nptr;
-          const unsigned long nm_bits = strtoul(slash + 1, &nptr, 10);
-          if (*nptr || 0 == nm_bits || nm_bits > (NULL != colon ? 128 : 32)) {
-            /*(also rejects (slash+1 == nptr) which results in nm_bits = 0)*/
-            fprintf(stderr, "invalid or missing netmask: %s\n", rvalue->ptr);
-            ctx->ok = 0;
-          }
-          else {
-            int rc;
-            buffer_string_set_length(rvalue, (size_t)(slash - rvalue->ptr)); /*(truncate)*/
-            rc = (NULL == colon)
-              ? http_request_host_normalize(rvalue, 0)
-              : configparser_remoteip_normalize_compat(rvalue);
-            buffer_append_string_len(rvalue, CONST_STR_LEN("/"));
-            buffer_append_int(rvalue, (int)nm_bits);
-            if (0 != rc) {
-              fprintf(stderr, "invalid IP addr: %s\n", rvalue->ptr);
-              ctx->ok = 0;
-            }
-          }
-        }
-        else {
-          int rc = (NULL == colon)
-            ? http_request_host_normalize(rvalue, 0)
-            : configparser_remoteip_normalize_compat(rvalue);
-          if (0 != rc) {
-            fprintf(stderr, "invalid IP addr: %s\n", rvalue->ptr);
-            ctx->ok = 0;
-          }
+        if (!config_remoteip_normalize(rvalue, tb)) {
+          fprintf(stderr, "invalid IP addr: %s\n", rvalue->ptr);
+          ctx->ok = 0;
         }
       }
       else if (COMP_SERVER_SOCKET == dc->comp) {
@@ -309,27 +265,10 @@ configparser_parse_condition(config_t * const ctx, const buffer * const obj_tag,
       }
 
       if (COMP_HTTP_REQUEST_HEADER == dc->comp) {
-        dc->ext = http_header_hkey_get(CONST_BUF_LEN(&dc->comp_tag));
+        dc->ext = http_header_hkey_get(BUF_PTR_LEN(&dc->comp_tag));
       }
 
-      buffer_copy_buffer(&dc->string, rvalue);
-
-      if (ctx->ok) switch(dc->cond) {
-      case CONFIG_COND_NE:
-      case CONFIG_COND_EQ:
-        break;
-      case CONFIG_COND_NOMATCH:
-      case CONFIG_COND_MATCH: {
-        if (!data_config_pcre_compile(dc)) {
-          ctx->ok = 0;
-        }
-        break;
-      }
-      default:
-        fprintf(stderr, "unknown condition for %s\n", dc->comp_key);
-        ctx->ok = 0;
-        break;
-      }
+      buffer_move(&dc->string, rvalue);
 
       if (ctx->ok)
         configparser_push(ctx, dc, 1);
@@ -343,14 +282,14 @@ configparser_parse_else_condition(config_t * const ctx)
 {
     data_config * const dc = data_config_init();
     dc->cond = CONFIG_COND_ELSE;
-    buffer_copy_buffer(&dc->key, &ctx->current->key);
-    buffer_append_string_len(&dc->key, CONST_STR_LEN(" / "));
-    buffer_append_string_len(&dc->key, CONST_STR_LEN("else_tmp_token"));
+    buffer_append_str2(&dc->key, BUF_PTR_LEN(&ctx->current->key),
+                                 CONST_STR_LEN(" / "
+                                               "else_tmp_token"));
     configparser_push(ctx, dc, 1);
 }
 
 
-#line 354 "./configparser.c"
+#line 293 "./configparser.c"
 /* Next is all token values, in a form suitable for use by makeheaders.
 ** This section will be null unless lemon is run with the -m switch.
 */
@@ -769,44 +708,44 @@ static void yy_destructor(YYCODETYPE yymajor, YYMINORTYPE *yypminor){
     case 24:
     case 25:
     case 26:
-#line 386 "../../src/configparser.y"
+#line 325 "../../src/configparser.y"
 { buffer_free((yypminor->yy0)); }
-#line 774 "./configparser.c"
+#line 713 "./configparser.c"
       break;
     case 36:
-#line 377 "../../src/configparser.y"
+#line 316 "../../src/configparser.y"
 { if ((yypminor->yy91)) (yypminor->yy91)->fn->free((yypminor->yy91)); }
-#line 779 "./configparser.c"
+#line 718 "./configparser.c"
       break;
     case 37:
-#line 378 "../../src/configparser.y"
+#line 317 "../../src/configparser.y"
 { if ((yypminor->yy91)) (yypminor->yy91)->fn->free((yypminor->yy91)); }
-#line 784 "./configparser.c"
+#line 723 "./configparser.c"
       break;
     case 38:
-#line 379 "../../src/configparser.y"
+#line 318 "../../src/configparser.y"
 { if ((yypminor->yy91)) (yypminor->yy91)->fn->free((yypminor->yy91)); }
-#line 789 "./configparser.c"
+#line 728 "./configparser.c"
       break;
     case 41:
-#line 380 "../../src/configparser.y"
+#line 319 "../../src/configparser.y"
 { array_free((yypminor->yy42)); }
-#line 794 "./configparser.c"
+#line 733 "./configparser.c"
       break;
     case 42:
-#line 381 "../../src/configparser.y"
+#line 320 "../../src/configparser.y"
 { array_free((yypminor->yy42)); }
-#line 799 "./configparser.c"
+#line 738 "./configparser.c"
       break;
     case 43:
-#line 382 "../../src/configparser.y"
+#line 321 "../../src/configparser.y"
 { buffer_free((yypminor->yy29)); }
-#line 804 "./configparser.c"
+#line 743 "./configparser.c"
       break;
     case 44:
-#line 383 "../../src/configparser.y"
+#line 322 "../../src/configparser.y"
 { buffer_free((yypminor->yy29)); }
-#line 809 "./configparser.c"
+#line 748 "./configparser.c"
       break;
     default:  break;   /* If no destructor action specified: do nothing */
   }
@@ -1083,9 +1022,9 @@ static void yy_reduce(
         /* No destructor defined for global */
         break;
       case 5:
-#line 359 "../../src/configparser.y"
+#line 298 "../../src/configparser.y"
 { yymsp[-1].minor.yy18 = NULL; }
-#line 1088 "./configparser.c"
+#line 1027 "./configparser.c"
   yy_destructor(1,&yymsp[0].minor);
         break;
       case 6:
@@ -1098,7 +1037,7 @@ static void yy_reduce(
   yy_destructor(1,&yymsp[0].minor);
         break;
       case 9:
-#line 388 "../../src/configparser.y"
+#line 327 "../../src/configparser.y"
 {
   if (ctx->ok) {
     buffer_copy_buffer(&yymsp[0].minor.yy91->key, yymsp[-2].minor.yy29);
@@ -1107,7 +1046,7 @@ static void yy_reduce(
           ctx->current->context_ndx,
           ctx->current->key.ptr, yymsp[-2].minor.yy29->ptr);
       ctx->ok = 0;
-    } else if (NULL == array_get_element_klen(ctx->current->value, CONST_BUF_LEN(&yymsp[0].minor.yy91->key))) {
+    } else if (NULL == array_get_element_klen(ctx->current->value, BUF_PTR_LEN(&yymsp[0].minor.yy91->key))) {
       array_insert_unique(ctx->current->value, yymsp[0].minor.yy91);
       yymsp[0].minor.yy91 = NULL;
     } else {
@@ -1122,11 +1061,11 @@ static void yy_reduce(
   if (yymsp[0].minor.yy91) yymsp[0].minor.yy91->fn->free(yymsp[0].minor.yy91);
   yymsp[0].minor.yy91 = NULL;
 }
-#line 1125 "./configparser.c"
+#line 1064 "./configparser.c"
   yy_destructor(2,&yymsp[-1].minor);
         break;
       case 10:
-#line 412 "../../src/configparser.y"
+#line 351 "../../src/configparser.y"
 {
   if (ctx->ok) {
     if (strncmp(yymsp[-2].minor.yy29->ptr, "env.", sizeof("env.") - 1) == 0) {
@@ -1145,11 +1084,11 @@ static void yy_reduce(
   if (yymsp[0].minor.yy91) yymsp[0].minor.yy91->fn->free(yymsp[0].minor.yy91);
   yymsp[0].minor.yy91 = NULL;
 }
-#line 1148 "./configparser.c"
+#line 1087 "./configparser.c"
   yy_destructor(3,&yymsp[-1].minor);
         break;
       case 11:
-#line 431 "../../src/configparser.y"
+#line 370 "../../src/configparser.y"
 {
   if (ctx->ok) {
     array *vars = ctx->current->value;
@@ -1160,7 +1099,7 @@ static void yy_reduce(
           ctx->current->context_ndx,
           ctx->current->key.ptr, yymsp[-2].minor.yy29->ptr);
       ctx->ok = 0;
-    } else if (NULL != (du = array_extract_element_klen(vars, CONST_BUF_LEN(yymsp[-2].minor.yy29))) || NULL != (du = configparser_get_variable(ctx, yymsp[-2].minor.yy29))) {
+    } else if (NULL != (du = array_extract_element_klen(vars, BUF_PTR_LEN(yymsp[-2].minor.yy29))) || NULL != (du = configparser_get_variable(ctx, yymsp[-2].minor.yy29))) {
       du = configparser_merge_data(du, yymsp[0].minor.yy91);
       if (NULL == du) {
         ctx->ok = 0;
@@ -1180,11 +1119,11 @@ static void yy_reduce(
   if (yymsp[0].minor.yy91) yymsp[0].minor.yy91->fn->free(yymsp[0].minor.yy91);
   yymsp[0].minor.yy91 = NULL;
 }
-#line 1183 "./configparser.c"
+#line 1122 "./configparser.c"
   yy_destructor(4,&yymsp[-1].minor);
         break;
       case 12:
-#line 462 "../../src/configparser.y"
+#line 401 "../../src/configparser.y"
 {
   if (strchr(yymsp[0].minor.yy0->ptr, '.') == NULL) {
     yygotominor.yy29 = buffer_init_string("var.");
@@ -1196,10 +1135,10 @@ static void yy_reduce(
   buffer_free(yymsp[0].minor.yy0);
   yymsp[0].minor.yy0 = NULL;
 }
-#line 1199 "./configparser.c"
+#line 1138 "./configparser.c"
         break;
       case 13:
-#line 474 "../../src/configparser.y"
+#line 413 "../../src/configparser.y"
 {
   yygotominor.yy91 = NULL;
   if (ctx->ok) {
@@ -1214,19 +1153,19 @@ static void yy_reduce(
   if (yymsp[0].minor.yy91) yymsp[0].minor.yy91->fn->free(yymsp[0].minor.yy91);
   yymsp[0].minor.yy91 = NULL;
 }
-#line 1217 "./configparser.c"
+#line 1156 "./configparser.c"
   yy_destructor(6,&yymsp[-1].minor);
         break;
       case 14:
-#line 489 "../../src/configparser.y"
+#line 428 "../../src/configparser.y"
 {
   yygotominor.yy91 = yymsp[0].minor.yy91;
   yymsp[0].minor.yy91 = NULL;
 }
-#line 1226 "./configparser.c"
+#line 1165 "./configparser.c"
         break;
       case 15:
-#line 494 "../../src/configparser.y"
+#line 433 "../../src/configparser.y"
 {
   yygotominor.yy91 = NULL;
   if (ctx->ok) {
@@ -1235,7 +1174,7 @@ static void yy_reduce(
 
       if (NULL != (env = getenv(yymsp[0].minor.yy29->ptr + 4))) {
         data_string *ds;
-        ds = data_string_init();
+        ds = array_data_string_init();
         buffer_append_string(&ds->value, env);
         yygotominor.yy91 = (data_unset *)ds;
       }
@@ -1251,24 +1190,24 @@ static void yy_reduce(
   buffer_free(yymsp[0].minor.yy29);
   yymsp[0].minor.yy29 = NULL;
 }
-#line 1254 "./configparser.c"
+#line 1193 "./configparser.c"
         break;
       case 16:
-#line 519 "../../src/configparser.y"
+#line 458 "../../src/configparser.y"
 {
-  yygotominor.yy91 = (data_unset *)data_string_init();
-  /* assumes data_string_init() result does not require swap and buffer_free()*/
+  yygotominor.yy91 = (data_unset *)array_data_string_init();
+  /* assumes array_data_string_init() result does not need swap, buffer_free()*/
   memcpy(&((data_string *)yygotominor.yy91)->value, yymsp[0].minor.yy0, sizeof(*yymsp[0].minor.yy0));
   free(yymsp[0].minor.yy0);
   yymsp[0].minor.yy0 = NULL;
 }
-#line 1265 "./configparser.c"
+#line 1204 "./configparser.c"
         break;
       case 17:
-#line 527 "../../src/configparser.y"
+#line 466 "../../src/configparser.y"
 {
   char *endptr;
-  yygotominor.yy91 = (data_unset *)data_integer_init();
+  yygotominor.yy91 = (data_unset *)array_data_integer_init();
   errno = 0;
   ((data_integer *)(yygotominor.yy91))->value = strtol(yymsp[0].minor.yy0->ptr, &endptr, 10);
   /* skip trailing whitespace */
@@ -1280,45 +1219,45 @@ static void yy_reduce(
   buffer_free(yymsp[0].minor.yy0);
   yymsp[0].minor.yy0 = NULL;
 }
-#line 1283 "./configparser.c"
+#line 1222 "./configparser.c"
         break;
       case 18:
-#line 541 "../../src/configparser.y"
+#line 480 "../../src/configparser.y"
 {
-  yygotominor.yy91 = (data_unset *)data_array_init();
-  /* assumes data_array_init() result does not require swap and array_free() */
+  yygotominor.yy91 = (data_unset *)array_data_array_init();
+  /* assumes array_data_array_init() result does not need swap, array_free() */
   memcpy(&((data_array *)(yygotominor.yy91))->value, yymsp[0].minor.yy42, sizeof(*yymsp[0].minor.yy42));
   free(yymsp[0].minor.yy42);
   yymsp[0].minor.yy42 = NULL;
 }
-#line 1294 "./configparser.c"
+#line 1233 "./configparser.c"
         break;
       case 19:
-#line 548 "../../src/configparser.y"
+#line 487 "../../src/configparser.y"
 {
   yygotominor.yy42 = array_init(8);
 }
-#line 1301 "./configparser.c"
+#line 1240 "./configparser.c"
   yy_destructor(9,&yymsp[-1].minor);
   yy_destructor(10,&yymsp[0].minor);
         break;
       case 20:
-#line 551 "../../src/configparser.y"
+#line 490 "../../src/configparser.y"
 {
   yygotominor.yy42 = yymsp[-1].minor.yy42;
   yymsp[-1].minor.yy42 = NULL;
 }
-#line 1311 "./configparser.c"
+#line 1250 "./configparser.c"
   yy_destructor(9,&yymsp[-2].minor);
   yy_destructor(10,&yymsp[0].minor);
         break;
       case 21:
-#line 556 "../../src/configparser.y"
+#line 495 "../../src/configparser.y"
 {
   yygotominor.yy42 = NULL;
   if (ctx->ok) {
-    if (buffer_is_empty(&yymsp[0].minor.yy91->key) ||
-        NULL == array_get_element_klen(yymsp[-2].minor.yy42, CONST_BUF_LEN(&yymsp[0].minor.yy91->key))) {
+    if (buffer_is_unset(&yymsp[0].minor.yy91->key) ||
+        NULL == array_get_element_klen(yymsp[-2].minor.yy42, BUF_PTR_LEN(&yymsp[0].minor.yy91->key))) {
       array_insert_unique(yymsp[-2].minor.yy42, yymsp[0].minor.yy91);
       yymsp[0].minor.yy91 = NULL;
     } else {
@@ -1335,20 +1274,20 @@ static void yy_reduce(
   if (yymsp[0].minor.yy91) yymsp[0].minor.yy91->fn->free(yymsp[0].minor.yy91);
   yymsp[0].minor.yy91 = NULL;
 }
-#line 1338 "./configparser.c"
+#line 1277 "./configparser.c"
   yy_destructor(11,&yymsp[-1].minor);
         break;
       case 22:
-#line 578 "../../src/configparser.y"
+#line 517 "../../src/configparser.y"
 {
   yygotominor.yy42 = yymsp[-1].minor.yy42;
   yymsp[-1].minor.yy42 = NULL;
 }
-#line 1347 "./configparser.c"
+#line 1286 "./configparser.c"
   yy_destructor(11,&yymsp[0].minor);
         break;
       case 23:
-#line 583 "../../src/configparser.y"
+#line 522 "../../src/configparser.y"
 {
   yygotominor.yy42 = NULL;
   if (ctx->ok) {
@@ -1359,18 +1298,18 @@ static void yy_reduce(
   if (yymsp[0].minor.yy91) yymsp[0].minor.yy91->fn->free(yymsp[0].minor.yy91);
   yymsp[0].minor.yy91 = NULL;
 }
-#line 1362 "./configparser.c"
+#line 1301 "./configparser.c"
         break;
       case 24:
-#line 594 "../../src/configparser.y"
+#line 533 "../../src/configparser.y"
 {
   yygotominor.yy91 = yymsp[0].minor.yy91;
   yymsp[0].minor.yy91 = NULL;
 }
-#line 1370 "./configparser.c"
+#line 1309 "./configparser.c"
         break;
       case 25:
-#line 598 "../../src/configparser.y"
+#line 537 "../../src/configparser.y"
 {
   yygotominor.yy91 = NULL;
   if (ctx->ok) {
@@ -1384,7 +1323,7 @@ static void yy_reduce(
   buffer_free(yymsp[-2].minor.yy29);
   yymsp[-2].minor.yy29 = NULL;
 }
-#line 1387 "./configparser.c"
+#line 1326 "./configparser.c"
   yy_destructor(12,&yymsp[-1].minor);
         break;
       case 26:
@@ -1393,31 +1332,31 @@ static void yy_reduce(
       case 27:
         break;
       case 28:
-#line 615 "../../src/configparser.y"
+#line 554 "../../src/configparser.y"
 {
   data_config *dc;
   dc = configparser_get_data_config(ctx->srv->config_context, CONST_STR_LEN("global"));
   force_assert(dc);
   configparser_push(ctx, dc, 0);
 }
-#line 1403 "./configparser.c"
+#line 1342 "./configparser.c"
   yy_destructor(13,&yymsp[0].minor);
         break;
       case 29:
-#line 622 "../../src/configparser.y"
+#line 561 "../../src/configparser.y"
 {
   force_assert(ctx->current);
   configparser_pop(ctx);
   force_assert(ctx->current);
 }
-#line 1413 "./configparser.c"
+#line 1352 "./configparser.c"
         /* No destructor defined for globalstart */
   yy_destructor(14,&yymsp[-2].minor);
         /* No destructor defined for metalines */
   yy_destructor(15,&yymsp[0].minor);
         break;
       case 30:
-#line 628 "../../src/configparser.y"
+#line 567 "../../src/configparser.y"
 {
   yygotominor.yy18 = NULL;
   if (ctx->ok) {
@@ -1436,12 +1375,12 @@ static void yy_reduce(
   yymsp[-3].minor.yy18 = NULL;
   yymsp[0].minor.yy18 = NULL;
 }
-#line 1439 "./configparser.c"
+#line 1378 "./configparser.c"
         /* No destructor defined for eols */
   yy_destructor(16,&yymsp[-1].minor);
         break;
       case 31:
-#line 647 "../../src/configparser.y"
+#line 586 "../../src/configparser.y"
 {
   yygotominor.yy18 = NULL;
   if (ctx->ok) {
@@ -1457,14 +1396,14 @@ static void yy_reduce(
   if (ctx->ok) {
     size_t pos;
     data_config *dc;
-    dc = (data_config *)array_extract_element_klen(ctx->all_configs, CONST_BUF_LEN(&yymsp[0].minor.yy18->key));
+    dc = (data_config *)array_extract_element_klen(ctx->all_configs, BUF_PTR_LEN(&yymsp[0].minor.yy18->key));
     force_assert(yymsp[0].minor.yy18 == dc);
     buffer_copy_buffer(&yymsp[0].minor.yy18->key, &yymsp[-3].minor.yy18->key);
     yymsp[0].minor.yy18->comp_key = yymsp[0].minor.yy18->key.ptr + (yymsp[-3].minor.yy18->comp_key - yymsp[-3].minor.yy18->key.ptr);
     yymsp[0].minor.yy18->comp = yymsp[-3].minor.yy18->comp;
     /*buffer_copy_buffer(&yymsp[0].minor.yy18->string, &yymsp[-3].minor.yy18->string);*/
     /* -2 for "==" and minus 3 for spaces and quotes around string (in key) */
-    pos = buffer_string_length(&yymsp[0].minor.yy18->key)-buffer_string_length(&yymsp[-3].minor.yy18->string)-5;
+    pos = buffer_clen(&yymsp[0].minor.yy18->key) - buffer_clen(&yymsp[-3].minor.yy18->string) - 5;
     switch(yymsp[-3].minor.yy18->cond) {
     case CONFIG_COND_NE:
       yymsp[0].minor.yy18->key.ptr[pos] = '='; /* opposite cond */
@@ -1486,7 +1425,7 @@ static void yy_reduce(
       force_assert(0);
     }
 
-    if (NULL == (dc = configparser_get_data_config(ctx->all_configs, CONST_BUF_LEN(&yymsp[0].minor.yy18->key)))) {
+    if (NULL == (dc = configparser_get_data_config(ctx->all_configs, BUF_PTR_LEN(&yymsp[0].minor.yy18->key)))) {
       /* re-insert into ctx->all_configs with new yymsp[0].minor.yy18->key */
       array_insert_unique(ctx->all_configs, (data_unset *)yymsp[0].minor.yy18);
       yymsp[0].minor.yy18->prev = yymsp[-3].minor.yy18;
@@ -1503,20 +1442,20 @@ static void yy_reduce(
   yymsp[-3].minor.yy18 = NULL;
   yymsp[0].minor.yy18 = NULL;
 }
-#line 1506 "./configparser.c"
+#line 1445 "./configparser.c"
         /* No destructor defined for eols */
   yy_destructor(16,&yymsp[-1].minor);
         break;
       case 32:
-#line 709 "../../src/configparser.y"
+#line 648 "../../src/configparser.y"
 {
   yygotominor.yy18 = yymsp[0].minor.yy18;
   yymsp[0].minor.yy18 = NULL;
 }
-#line 1516 "./configparser.c"
+#line 1455 "./configparser.c"
         break;
       case 33:
-#line 714 "../../src/configparser.y"
+#line 653 "../../src/configparser.y"
 {
   yygotominor.yy18 = NULL;
   if (ctx->ok) {
@@ -1530,14 +1469,14 @@ static void yy_reduce(
     yygotominor.yy18 = cur;
   }
 }
-#line 1533 "./configparser.c"
+#line 1472 "./configparser.c"
         /* No destructor defined for context */
   yy_destructor(14,&yymsp[-2].minor);
         /* No destructor defined for metalines */
   yy_destructor(15,&yymsp[0].minor);
         break;
       case 34:
-#line 728 "../../src/configparser.y"
+#line 667 "../../src/configparser.y"
 {
   yygotominor.yy18 = NULL;
   if (ctx->ok) {
@@ -1551,14 +1490,14 @@ static void yy_reduce(
     yygotominor.yy18 = cur;
   }
 }
-#line 1554 "./configparser.c"
+#line 1493 "./configparser.c"
         /* No destructor defined for context_else */
   yy_destructor(14,&yymsp[-2].minor);
         /* No destructor defined for metalines */
   yy_destructor(15,&yymsp[0].minor);
         break;
       case 35:
-#line 742 "../../src/configparser.y"
+#line 681 "../../src/configparser.y"
 {
 
   if (ctx->ok && yymsp[0].minor.yy91->type != TYPE_STRING) {
@@ -1577,54 +1516,54 @@ static void yy_reduce(
   yymsp[0].minor.yy91->fn->free(yymsp[0].minor.yy91);
   yymsp[0].minor.yy91 = NULL;
 }
-#line 1580 "./configparser.c"
+#line 1519 "./configparser.c"
   yy_destructor(17,&yymsp[-6].minor);
   yy_destructor(19,&yymsp[-4].minor);
   yy_destructor(20,&yymsp[-2].minor);
         break;
       case 36:
-#line 761 "../../src/configparser.y"
+#line 700 "../../src/configparser.y"
 {
   if (ctx->ok) {
     configparser_parse_else_condition(ctx);
   }
 }
-#line 1592 "./configparser.c"
+#line 1531 "./configparser.c"
         break;
       case 37:
-#line 767 "../../src/configparser.y"
+#line 706 "../../src/configparser.y"
 {
   yygotominor.yy53 = CONFIG_COND_EQ;
 }
-#line 1599 "./configparser.c"
+#line 1538 "./configparser.c"
   yy_destructor(21,&yymsp[0].minor);
         break;
       case 38:
-#line 770 "../../src/configparser.y"
+#line 709 "../../src/configparser.y"
 {
   yygotominor.yy53 = CONFIG_COND_MATCH;
 }
-#line 1607 "./configparser.c"
+#line 1546 "./configparser.c"
   yy_destructor(22,&yymsp[0].minor);
         break;
       case 39:
-#line 773 "../../src/configparser.y"
+#line 712 "../../src/configparser.y"
 {
   yygotominor.yy53 = CONFIG_COND_NE;
 }
-#line 1615 "./configparser.c"
+#line 1554 "./configparser.c"
   yy_destructor(23,&yymsp[0].minor);
         break;
       case 40:
-#line 776 "../../src/configparser.y"
+#line 715 "../../src/configparser.y"
 {
   yygotominor.yy53 = CONFIG_COND_NOMATCH;
 }
-#line 1623 "./configparser.c"
+#line 1562 "./configparser.c"
   yy_destructor(24,&yymsp[0].minor);
         break;
       case 41:
-#line 780 "../../src/configparser.y"
+#line 719 "../../src/configparser.y"
 {
   yygotominor.yy29 = NULL;
   if (ctx->ok) {
@@ -1641,10 +1580,10 @@ static void yy_reduce(
   if (yymsp[0].minor.yy91) yymsp[0].minor.yy91->fn->free(yymsp[0].minor.yy91);
   yymsp[0].minor.yy91 = NULL;
 }
-#line 1644 "./configparser.c"
+#line 1583 "./configparser.c"
         break;
       case 42:
-#line 797 "../../src/configparser.y"
+#line 736 "../../src/configparser.y"
 {
   if (ctx->ok) {
     if (0 != config_parse_file(ctx->srv, ctx, yymsp[0].minor.yy29->ptr)) {
@@ -1654,11 +1593,11 @@ static void yy_reduce(
   buffer_free(yymsp[0].minor.yy29);
   yymsp[0].minor.yy29 = NULL;
 }
-#line 1657 "./configparser.c"
+#line 1596 "./configparser.c"
   yy_destructor(25,&yymsp[-1].minor);
         break;
       case 43:
-#line 807 "../../src/configparser.y"
+#line 746 "../../src/configparser.y"
 {
   if (ctx->ok) {
     if (0 != config_parse_cmd(ctx->srv, ctx, yymsp[0].minor.yy29->ptr)) {
@@ -1668,7 +1607,7 @@ static void yy_reduce(
   buffer_free(yymsp[0].minor.yy29);
   yymsp[0].minor.yy29 = NULL;
 }
-#line 1671 "./configparser.c"
+#line 1610 "./configparser.c"
   yy_destructor(26,&yymsp[-1].minor);
         break;
   };
@@ -1698,11 +1637,11 @@ static void yy_parse_failed(
   while( yypParser->yyidx>=0 ) yy_pop_parser_stack(yypParser);
   /* Here code is inserted which will be executed whenever the
   ** parser fails */
-#line 350 "../../src/configparser.y"
+#line 289 "../../src/configparser.y"
 
   ctx->ok = 0;
 
-#line 1705 "./configparser.c"
+#line 1644 "./configparser.c"
   configparserARG_STORE; /* Suppress warning about unused %extra_argument variable */
 }
 
