@@ -290,6 +290,12 @@ SETDEFAULTS_FUNC(mod_authn_mysql_set_defaults) {
             mod_authn_mysql_merge_config(&p->defaults, cpv);
     }
 
+    log_error(srv->errh, __FILE__, __LINE__,
+      "Warning: mod_%s is deprecated "
+      "and will be removed from a future lighttpd release in early 2022. "
+      "https://wiki.lighttpd.net/Docs_ModAuth#mysql-mod_authn_mysql-since-lighttpd-1442",
+      p->self->name);
+
     return HANDLER_GO_ON;
 }
 
@@ -299,7 +305,7 @@ static int mod_authn_mysql_password_cmp(const char *userpw, unsigned long userpw
         char *crypted = crypt(reqpw, userpw);
         size_t crypwlen = (NULL != crypted) ? strlen(crypted) : 0;
         int rc = (crypwlen == userpwlen) ? memcmp(crypted, userpw, crypwlen) : -1;
-        if (crypwlen) ck_memzero(crypted, crypwlen);
+        if (crypwlen >= 13) ck_memzero(crypted, crypwlen);
         return rc;
     }
     else
@@ -471,6 +477,7 @@ static handler_t mod_authn_mysql_basic(request_st * const r, void *p_d, const ht
     ai.ulen     = buffer_clen(username);
     ai.realm    = require->realm->ptr;
     ai.rlen     = buffer_clen(require->realm);
+    ai.userhash = 0;
     rc = mod_authn_mysql_query(r, p_d, &ai, pw);
     if (HANDLER_GO_ON != rc) return rc;
     return http_auth_match_rules(require, username->ptr, NULL, NULL)
