@@ -354,42 +354,7 @@ static void fastcgi_get_packet_body(buffer * const b, handler_ctx * const hctx,
 }
 
 __attribute_cold__
-__attribute_noinline__
-static int
-mod_fastcgi_chunk_decode_transfer_cqlen (request_st * const r, chunkqueue * const src, const unsigned int len)
-{
-    if (0 == len) return 0;
-
-    /* specialized for mod_fastcgi to decode chunked encoding;
-     * FastCGI packet data is all type MEM_CHUNK
-     * entire src cq is processed, minus packet.padding at end
-     * (This extra work can be avoided if FastCGI backend does not send
-     *  Transfer-Encoding: chunked, which FastCGI is not supposed to do) */
-    uint32_t remain = len, wr;
-    for (const chunk *c = src->first; c && remain; c = c->next, remain -= wr) {
-        /*assert(c->type == MEM_CHUNK);*/
-        wr = buffer_clen(c->mem) - c->offset;
-        if (wr > remain) wr = remain;
-        if (0 != http_chunk_decode_append_mem(r, c->mem->ptr+c->offset, wr))
-            return -1;
-    }
-    chunkqueue_mark_written(src, len);
-    return 0;
-}
-
-static int
-mod_fastcgi_transfer_cqlen (request_st * const r, chunkqueue * const src, const unsigned int len)
-{
-    return (!r->resp_decode_chunked)
-      ? http_chunk_transfer_cqlen(r, src, len)
-      : mod_fastcgi_chunk_decode_transfer_cqlen(r, src, len);
-}
-
-static handler_t fcgi_recv_parse(request_st * const r, struct http_response_opts_t *opts, buffer *b, size_t n) {
-	handler_ctx *hctx = (handler_ctx *)opts->pdata;
-	int fin = 0;
-
-	if (0 == n) {
+static handler_t fcgi_recv_0(const request_st * const r, const handler_ctx * const hctx) {
 		if (-1 == hctx->request_id) return HANDLER_FINISHED; /*(flag request ended)*/
 		if (!(fdevent_fdnode_interest(hctx->fdn) & FDEVENT_IN)
 		    && !(r->conf.stream_response_body & FDEVENT_STREAM_RESPONSE_POLLRDHUP))
@@ -400,17 +365,16 @@ static handler_t fcgi_recv_parse(request_st * const r, struct http_response_opts
 		  hctx->proc->pid, hctx->proc->connection_name->ptr);
 
 		return HANDLER_ERROR;
-	}
-
-	chunkqueue_append_buffer(hctx->rb, b);
+}
 
+static handler_t fcgi_recv_parse_loop(request_st * const r, handler_ctx * const hctx) {
 	/*
 	 * parse the fastcgi packets and forward the content to the write-queue
 	 *
 	 */
-	while (fin == 0) {
-		fastcgi_response_packet packet;
-
+	fastcgi_response_packet packet;
+	int fin = 0;
+	do {
 		/* check if we have at least one packet */
 		if (0 != fastcgi_get_packet(hctx, &packet)) {
 			/* no full packet */
@@ -445,17 +409,33 @@ static handler_t fcgi_recv_parse(request_st * const r, struct http_response_opts
 					 (r->http_status == 0 || r->http_status == 200)) {
 					/* authorizer approved request; ignore the content here */
 					hctx->send_content_body = 0;
-					opts->authorizer |= /*(save response streaming flags)*/
+					hctx->opts.authorizer |= /*(save response streaming flags)*/
 					  (r->conf.stream_response_body
 					   & (FDEVENT_STREAM_RESPONSE
 					     |FDEVENT_STREAM_RESPONSE_BUFMIN)) << 1;
 					r->conf.stream_response_body &=
 					  ~(FDEVENT_STREAM_RESPONSE|FDEVENT_STREAM_RESPONSE_BUFMIN);
 				}
+			  #if 0
+				else if ((r->conf.stream_response_body &
+				           (FDEVENT_STREAM_RESPONSE|FDEVENT_STREAM_RESPONSE_BUFMIN))
+				         && (   r->http_status == 204
+				             || r->http_status == 205
+				             || r->http_status == 304
+				             || r->http_method == HTTP_METHOD_HEAD)) {
+					/* disable streaming to wait for backend protocol to signal
+					 * end of response (prevent http_response_write_prepare()
+					 * from short-circuiting and finishing responses without
+					 * response body) */
+					r->conf.stream_response_body &=
+					  ~(FDEVENT_STREAM_RESPONSE|FDEVENT_STREAM_RESPONSE_BUFMIN);
+				}
+			  #endif
 			} else if (hctx->send_content_body) {
-				if (0 != mod_fastcgi_transfer_cqlen(r, hctx->rb, packet.len - packet.padding)) {
+				if (0 != http_response_transfer_cqlen(r, hctx->rb, (size_t)(packet.len - packet.padding))) {
 					/* error writing to tempfile;
 					 * truncate response or send 500 if nothing sent yet */
+					hctx->send_content_body = 0;
 					fin = 1;
 				}
 				if (packet.padding) chunkqueue_mark_written(hctx->rb, packet.padding);
@@ -482,11 +462,18 @@ static handler_t fcgi_recv_parse(request_st * const r, struct http_response_opts
 			chunkqueue_mark_written(hctx->rb, packet.len);
 			break;
 		}
-	}
+	} while (0 == fin);
 
 	return 0 == fin ? HANDLER_GO_ON : HANDLER_FINISHED;
 }
 
+static handler_t fcgi_recv_parse(request_st * const r, struct http_response_opts_t *opts, buffer *b, size_t n) {
+    handler_ctx * const hctx = (handler_ctx *)opts->pdata;
+    if (0 == n) return fcgi_recv_0(r, hctx);
+    chunkqueue_append_buffer(hctx->rb, b);
+    return fcgi_recv_parse_loop(r, hctx);
+}
+
 static handler_t fcgi_check_extension(request_st * const r, void *p_d, int uri_path_handler) {
 	plugin_data *p = p_d;
 	handler_t rc;
