@@ -68,6 +68,7 @@ static void* mod_mysql_vhost_connection_data(request_st * const r, void *p_d)
 
 	if (c) return c;
 	c = calloc(1, sizeof(*c));
+	force_assert(c);
 
 	c->server_name = buffer_init();
 	c->document_root = buffer_init();
@@ -262,6 +263,12 @@ SETDEFAULTS_FUNC(mod_mysql_vhost_set_defaults) {
             mod_mysql_vhost_merge_config(&p->defaults, cpv);
     }
 
+    log_error(srv->errh, __FILE__, __LINE__,
+      "Warning: mod_%s is deprecated "
+      "and will be removed from a future lighttpd release in early 2022. "
+      "https://wiki.lighttpd.net/Docs_ConfigurationOptions#Deprecated",
+      p->self->name);
+
     return HANDLER_GO_ON;
 }
 
