@@ -18,6 +18,11 @@
 #include "request.h"
 #include "response.h"
 
+#ifdef HAVE_PCRE2_H
+#define PCRE2_CODE_UNIT_WIDTH 8
+#include <pcre2.h>
+#endif
+
 
 static const request_config *request_config_defaults;
 
@@ -60,11 +65,13 @@ request_init_data (request_st * const r, connection * const con, server * const
     r->cond_cache = calloc(srv->config_context->used, sizeof(cond_cache_t));
     force_assert(NULL != r->cond_cache);
 
-  #ifdef HAVE_PCRE_H
-    if (srv->config_context->used > 1) {/*(save 128b per con if no conditions)*/
-        r->cond_match =
-          calloc(srv->config_context->used, sizeof(cond_match_t));
+  #ifdef HAVE_PCRE
+    if (srv->config_captures) {
+        r->cond_captures = srv->config_captures;
+        r->cond_match = calloc(srv->config_captures, sizeof(cond_match_t *));
         force_assert(NULL != r->cond_match);
+        r->cond_match_data = calloc(srv->config_captures, sizeof(cond_match_t));
+        force_assert(NULL != r->cond_match_data);
     }
   #endif
 
@@ -224,7 +231,21 @@ request_free_data (request_st * const r)
 
     free(r->plugin_ctx);
     free(r->cond_cache);
-    free(r->cond_match);
+  #ifdef HAVE_PCRE
+    if (r->cond_match_data) {
+        for (int i = 0, used = r->cond_captures; i < used; ++i) {
+          #ifdef HAVE_PCRE2_H
+            if (r->cond_match_data[i].match_data)
+                pcre2_match_data_free(r->cond_match_data[i].match_data);
+          #else /* HAVE_PCRE_H */
+            if (r->cond_match_data[i].matches)
+                free(r->cond_match_data[i].matches);
+          #endif
+        }
+        free(r->cond_match_data);
+        free(r->cond_match);
+    }
+  #endif
 
     /* note: r is not zeroed here and r is not freed here */
 }
